from datetime import datetime, timedelta, date, time
import calendar
from django.db.models import Q
from django.utils import timezone

import logging

logger = logging.getLogger(__name__)


def daterange(start, end, step=timedelta(1)):
    if end is None:
        yield start
    else:
        curr = start
        while curr <= end:
            yield curr
            curr += step


class DayInCalendar:
    midnight = time(23, 59, 59)

    def __init__(self, d, on_requested_interval=True):
        self.date = d
        now = date.today()
        self.week = d.isocalendar()[1]

        self.in_past = d < now
        self.today = d == now
        self.events = []
        self.on_requested_interval = on_requested_interval

        self.events_by_category = {}

    def is_in_past(self):
        return self.in_past

    def is_today(self):
        return self.today

    def is_ancestor_uuid_event_from_other(self, event):
        for e in self.events:
            if event.is_ancestor_by_uuid(e):
                return True
        return False

    def remove_event_with_ancestor_uuid_if_exists(self, event):
        removed = False
        for i, e in enumerate(self.events):
            if e.is_ancestor_by_uuid(event):
                # remove e from events_by_category
                for k, v in self.events_by_category.items():
                    if e in v:
                        self.events_by_category[k].remove(e)
                # remove e from the events
                del self.events[i]
                removed = True

        if removed:
            # remove empty events_by_category
            self.events_by_category = dict(
                [(k, v) for k, v in self.events_by_category.items() if len(v) > 0]
            )

    def add_event(self, event):
        if event.contains_date(self.date):
            if self.is_ancestor_uuid_event_from_other(event):
                # we do not add a generic event if a specific is already present
                pass
            else:
                self.remove_event_with_ancestor_uuid_if_exists(event)
                self._add_event_internal(event)

    def _add_event_internal(self, event):
        self.events.append(event)
        if event.category is None:
            if "" not in self.events_by_category:
                self.events_by_category[""] = []
            self.events_by_category[""].append(event)
        else:
            if event.category.name not in self.events_by_category:
                self.events_by_category[event.category.name] = []
            self.events_by_category[event.category.name].append(event)

    def filter_events(self):
        self.events.sort(
            key=lambda e: DayInCalendar.midnight
            if e.start_time is None
            else e.start_time
        )


class CalendarList:
    def __init__(self, firstdate, lastdate, filter=None, exact=False):
        self.firstdate = firstdate
        self.lastdate = lastdate
        self.now = date.today()
        self.filter = filter

        if exact:
            self.c_firstdate = self.firstdate
            self.c_lastdate = self.lastdate
        else:
            # start the first day of the first week
            self.c_firstdate = firstdate + timedelta(days=-firstdate.weekday())
            # end the last day of the last week
            self.c_lastdate = lastdate + timedelta(days=6 - lastdate.weekday())

        self.calendar_days = None


    def build_internal(self):
        # create a list of DayInCalendars
        self.create_calendar_days()

        # fill DayInCalendars with events
        self.fill_calendar_days()

        # finally, sort each DayInCalendar
        for i, c in self.calendar_days.items():
            c.filter_events()

    def get_calendar_days(self):
        if self.calendar_days is None:
            self.build_internal()
        
        return self.calendar_days

    def today_in_calendar(self):
        return self.firstdate <= self.now and self.lastdate >= self.now

    def all_in_past(self):
        return self.lastdate < self.now

    def fill_calendar_days(self):
        if self.filter is None:
            from .models import Event

            qs = Event.objects.all()
        else:
            qs = self.filter.qs
        startdatetime = timezone.make_aware(datetime.combine(self.c_firstdate, time.min), timezone.get_default_timezone())
        lastdatetime = timezone.make_aware(datetime.combine(self.c_lastdate, time.max), timezone.get_default_timezone())
        self.events = qs.filter(
            (Q(recurrence_dtend__isnull=True) & Q(recurrence_dtstart__lte=lastdatetime))
            | (
                Q(recurrence_dtend__isnull=False)
                & ~(
                    Q(recurrence_dtstart__gt=lastdatetime)
                    | Q(recurrence_dtend__lt=startdatetime)
                )
            )
        ).order_by("start_time")

        firstdate = datetime.fromordinal(self.c_firstdate.toordinal())
        if firstdate.tzinfo is None or firstdate.tzinfo.utcoffset(firstdate) is None:
            firstdate = timezone.make_aware(firstdate, timezone.get_default_timezone())

        lastdate = datetime.fromordinal(self.c_lastdate.toordinal())
        if lastdate.tzinfo is None or lastdate.tzinfo.utcoffset(lastdate) is None:
            lastdate = timezone.make_aware(lastdate, timezone.get_default_timezone())

        for e in self.events:
            for e_rec in e.get_recurrences_between(firstdate, lastdate):
                for d in daterange(e_rec.start_day, e_rec.end_day):
                    if d.__str__() in self.calendar_days:
                        self.calendar_days[d.__str__()].add_event(e_rec)

    def create_calendar_days(self):
        # create daylist
        self.calendar_days = {}
        for d in daterange(self.c_firstdate, self.c_lastdate):
            self.calendar_days[d.strftime("%Y-%m-%d")] = DayInCalendar(
                d, d >= self.firstdate and d <= self.lastdate
            )

    def is_single_week(self):
        return hasattr(self, "week")

    def is_full_month(self):
        return hasattr(self, "month")

    def calendar_days_list(self):
        return list(self.get_calendar_days().values())

    def export_to_ics(self):
        from .models import Event
        events = [event for day in self.get_calendar_days().values() for event in day.events]
        return Event.export_to_ics(events)


class CalendarMonth(CalendarList):
    def __init__(self, year, month, filter):
        self.year = year
        self.month = month
        r = calendar.monthrange(year, month)

        first = date(year, month, 1)
        last = date(year, month, r[1])

        super().__init__(first, last, filter)

    def get_month_name(self):
        return self.firstdate.strftime("%B")

    def next_month(self):
        return self.lastdate + timedelta(days=7)

    def previous_month(self):
        return self.firstdate + timedelta(days=-7)


class CalendarWeek(CalendarList):
    def __init__(self, year, week, filter):
        self.year = year
        self.week = week

        first = date.fromisocalendar(self.year, self.week, 1)
        last = date.fromisocalendar(self.year, self.week, 7)

        super().__init__(first, last, filter)

    def next_week(self):
        return self.firstdate + timedelta(days=7)

    def previous_week(self):
        return self.firstdate + timedelta(days=-7)


class CalendarDay(CalendarList):
    def __init__(self, date, filter=None):
        super().__init__(date, date, filter, exact=True)

    def get_events(self):
        return self.calendar_days_list()[0].events
