from abc import abstractmethod
from urllib.parse import urlparse
from urllib.parse import parse_qs

import logging

logger = logging.getLogger(__name__)


from .extractor import *
from django.utils.translation import gettext_lazy as _
from dateutil import parser
import datetime


class GGCalendar:
    def __init__(self, url):
        self.url = url
        self.extract_info()

    def filter_keys(params):
        result = {}

        for k, v in params.items():
            if k.startswith('e[0]'):
                result[k.replace('e[0][', '')[:-1]] = v
            else:
                result[k] = v

        return result

    def is_valid_event(self):
        return self.start_day is not None and self.title is not None

    def extract_info(self):
        parsed_url = urlparse(self.url.replace("#", "%23"))
        params = parse_qs(parsed_url.query)

        params = GGCalendar.filter_keys(params)

        self.location = params["location"][0] if "location" in params else None
        self.title = params["text"][0] if "text" in params else params["title"][0] if "title" in params else None
        self.description = params["description"][0] if "description" in params else None
        self.location = params["location"][0] if "location" in params else None
        if "dates" in params:
            dates = [x.replace(" ", "+") for x in params["dates"][0].split("/")]
            if len(dates) > 0:
                date = parser.parse(dates[0])
                self.start_day = date.date()
                self.start_time = date.time()
                if len(dates) == 2:
                    date = parser.parse(dates[1])
                    self.end_day = date.date()
                    self.end_time = date.time()
                else:
                    self.end_day = None
                    self.end_time = None
        elif "date_start" in params:
            date = parser.parse(params["date_start"][0])
            self.start_day = date.date()
            self.start_time = date.time()
            if "date_end" in params:
                dateend = parser.parse(params["date_end"][0])
                if dateend != date:
                    self.end_day = dateend.date()
                    self.end_time = dateend.time()
                else:
                    self.end_day = None
                    self.end_time = None
                    if self.start_time == datetime.time(0):
                        self.start_time = None

            else:
                self.end_day = None
                self.end_time = None
        else:
            raise Exception("Unable to find a date in google calendar URL")
            self.start_day = None
            self.start_time = None
            self.end_day = None
            self.end_time = None


# A class to extract events from URL with two steps:
# - first build a list of urls where the events will be found
# - then for each document downloaded from these urls, build the events
# This class is an abstract class
class TwoStepsExtractor(Extractor):

    def __init__(self):
        super().__init__()
        self.event_urls = None
        self.event_properties = {}

    def clean_url(url):
        return url

    def add_event_url(self, url):
        if url in self.event_urls:
            return False
        else:
            self.event_urls.append(url)
            return True

    def add_event_start_day(self, url, start_day):
        if url not in self.event_properties:
            self.event_properties[url] = {}
        self.event_properties[url]["start_day"] = start_day

    def add_event_start_time(self, url, start_time):
        if url not in self.event_properties:
            self.event_properties[url] = {}
        self.event_properties[url]["start_time"] = start_time

    def add_event_title(self, url, title):
        if url not in self.event_properties:
            self.event_properties[url] = {}
        self.event_properties[url]["title"] = title

    def add_event_tag(self, url, tag):
        if url not in self.event_properties:
            self.event_properties[url] = {}
        if "tags" not in self.event_properties[url]:
            self.event_properties[url]["tags"] = []
        self.event_properties[url]["tags"].append(tag)

    def add_event_category(self, url, cat):
        if url not in self.event_properties:
            self.event_properties[url] = {}
        self.event_properties[url]["category"] = cat

    def add_event_location(self, url, loc):
        if url not in self.event_properties:
            self.event_properties[url] = {}
        self.event_properties[url]["location"] = loc

    def add_event_with_props(
        self,
        event_url,
        title,
        category,
        start_day,
        location,
        description,
        tags,
        uuids,
        recurrences=None,
        url_human=None,
        start_time=None,
        end_day=None,
        end_time=None,
        last_modified=None,
        published=False,
        image=None,
        image_alt=None,
    ):
        if event_url in self.event_properties:
            if "tags" in self.event_properties[event_url]:
                tags = tags + self.event_properties[event_url]["tags"]
            if "start_day" in self.event_properties[event_url]:
                start_day = self.event_properties[event_url]["start_day"]
            if "start_time" in self.event_properties[event_url]:
                start_time = self.event_properties[event_url]["start_time"]
            if "title" in self.event_properties[event_url]:
                title = self.event_properties[event_url]["title"]
            if "category" in self.event_properties[event_url]:
                category = self.event_properties[event_url]["category"]
            if "location" in self.event_properties[event_url]:
                location = self.event_properties[event_url]["location"]

        self.add_event(
            title,
            category,
            start_day,
            location,
            description,
            tags,
            uuids,
            recurrences,
            url_human,
            start_time,
            end_day,
            end_time,
            last_modified,
            published,
            image,
            image_alt,
        )

    @abstractmethod
    def build_event_url_list(self, content):
        pass

    @abstractmethod
    def add_event_from_content(
        self,
        event_content,
        event_url,
        url_human=None,
        default_values=None,
        published=False,
    ):
        pass

    def extract(
        self,
        content,
        url,
        url_human=None,
        default_values=None,
        published=False,
        only_future=True,
        ignore_404=True
    ):
        self.only_future = only_future
        self.now = datetime.datetime.now().date()
        self.set_header(url)
        self.clear_events()

        self.url = url
        self.event_urls = []
        self.event_properties.clear()

        # first build the event list
        self.build_event_url_list(content)

        if self.event_urls is None:
            raise Exception("Unable to find the event list from the main document")

        if self.downloader is None:
            raise Exception("The downloader is not defined")

        # then process each element of the list
        for i, event_url in enumerate(self.event_urls):
            # first download the content associated with this link
            content_event = self.downloader.get_content(event_url)
            if content_event is None:
                msg = "Cannot extract event from url {}".format(event_url)
                if ignore_404:
                    logger.error(msg)
                else:
                    print("go")
                    raise Exception(msg)
            else:
                # then extract event information from this html document
                self.add_event_from_content(
                    content_event, event_url, url_human, default_values, published
                )

        return self.get_structure()
